import java.util.*;

public class GenMethods {
	
	static double square(Number x) {
		double y = x.doubleValue();
		return 1.0 * y*y;
	}
	
	static Number sum(List<? extends Number> numbers) {
		double sum = 0;
		for (Number n: numbers)
			sum += n.doubleValue();
		return sum;
	}
	
	static <E> void uniqueNumbers(List<? extends E> numbers, List<? super E> uniques) {
		for (E n: numbers) 
			if (!uniques.contains(n)) uniques.add(n);
	}
	
	static <E> void listToListCopy(List<E> numbersIn, List<? super E> numbersOut) {
		// Iterator<? extends E> it = numbersIn.iterator(); 
		// while (it.hasNext()) {
		// 	addElement(numbersOut, it.next());
		// }
		for (E n: numbersIn)
			numbersOut.add(n);
	}
	
	// private static <E> void addElement(List<? super E> list, E el) {
	// 	list.add(el);
	// }
	
	public static void main(String[] args) {
		double y = 10.0;
		int x = 10;
		
		// producer extends rule example: the method sum()
		System.out.printf("squared %d is %.2f, and squared %.2f is %.2f%n", 
						x, square(x), y, square(y));
		List<Double> dnumbers = Arrays.asList(new Double[] {1.0, 1.0, 3.1, 5.2, 9.5});
		System.out.println(sum(dnumbers));
		List<Integer> inumbers = Arrays.asList(new Integer[] {1, 1, 3, 5, 9});
		System.out.println(sum(inumbers));
		
		// consumer super rule example: the method uniqueNumbers()
		System.out.printf("original integers %s%n", inumbers);
		System.out.printf("original doubles %s%n", dnumbers);
		List<Number> uniques = new ArrayList<>();
		uniqueNumbers(inumbers, uniques);
		System.out.printf("unique numbers %s%n", uniques);
		GenMethods.<Number>uniqueNumbers(dnumbers, uniques); // explicit TP is redundant
		System.out.printf("more unique numbers %s%n", uniques);

		// simultanneous PECS rule example: the method listToListCopy()
		List<Integer> intList = new ArrayList<>();
		List<Double> doubleList = new ArrayList<>();
		List<Number> numberList = new ArrayList<>();
		listToListCopy(inumbers, intList); // E is Integer
		System.out.printf("integers copied to integers %s%n", intList);
		listToListCopy(dnumbers, doubleList); // E is Double
		System.out.printf("doubles copied to doubles %s%n", doubleList);
		listToListCopy(inumbers,numberList); // explicit TP is redundant
		System.out.printf("integers copied to numbers %s%n", numberList);
		listToListCopy(dnumbers,numberList); // explicit TP is redundant
		System.out.printf("doubles added to numbers to which integres were copied %s%n", numberList);
	}
	
}